from allophant.datasets.speech_corpus import MultilingualCorpus, MultilingualSplits, AudioInfo
from allophant.datasets.ucla_phonetic_corpus import UCLAPhoneticCorpus
from allophant.datasets.mozilla_common_voice import CommonVoiceCorpus, CommonVoiceSplits


__all__ = [
    "MultilingualCorpus",
    "MultilingualSplits",
    "AudioInfo",
    "UCLAPhoneticCorpus",
    "CommonVoiceSplits",
    "CommonVoiceCorpus",
]
